from __future__ import annotations

import typing as t

import click

from gcs_cli.helpers import requires_login
from gcs_cli.parsing import (
    EXPLICIT_NULL,
    auth_policy_create_and_update_params,
    auth_policy_id_arg,
    globus_command,
)
from gcs_cli.sdk_client.auth_client import MissingConsent
from gcs_cli.sdk_client.data import AuthPolicyDocument
from gcs_cli.services.auth import get_client

if t.TYPE_CHECKING:
    import uuid

    from gcs_cli.parsing.types import _ExplicitNullClass


@globus_command(
    "update",
    short_help="Update an existing authentication policy",
    disable_options=["explicit_host", "format"],
)
@requires_login
@auth_policy_id_arg
@click.option(
    "--description",
    metavar="DESCRIPTION",
    type=click.STRING,
    help="Description for the authentication policy.",
)
@click.option(
    "--display-name",
    metavar="DISPLAY_NAME",
    type=click.STRING,
    help="Display name for the authentication policy.",
)
@auth_policy_create_and_update_params(create=False)
def auth_policy_update(
    auth_policy_id: uuid.UUID,
    include: t.Iterable[str],
    exclude: t.Iterable[str],
    description: str | None,
    display_name: str | None,
    **kwargs: t.Mapping[str, t.Any],
) -> None:
    client = get_client()

    def _convert_domain_tuple(
        domain_tuple: t.Iterable[str],
    ) -> t.List[str] | _ExplicitNullClass | None:
        # Remove any duplicates
        domain_list = list(set(domain_tuple))
        # Handle an explicit unsetting of the domain constraint
        if domain_list == [""]:
            return EXPLICIT_NULL
        # Remove any superfluous ''
        domain_list = list(filter(lambda x: x != "", domain_list))
        # Handle no values given
        if domain_list == []:
            return None
        return domain_list

    auth_policy_doc = AuthPolicyDocument(
        project_id=None,
        display_name=display_name,
        description=description,
        domain_constraints_include=_convert_domain_tuple(include),
        domain_constraints_exclude=_convert_domain_tuple(exclude),
    )

    try:
        client.update_auth_policy(auth_policy_id, auth_policy_doc)
    except MissingConsent as e:
        raise click.ClickException(
            "Missing consent for the policy API. Please run the following command:\n"
            f"    globus-connect-server session consent {e.scope}"
        )
    click.echo(f"Updated authentication policy {auth_policy_id}")
