from __future__ import annotations

import re
import typing as t
import uuid
from datetime import date

import click

from globus_sdk.paging import Paginator

from gcs_cli.commands.collection._get_last_access import get_last_access
from gcs_cli.helpers import default_identity_id_resolver, requires_login
from gcs_cli.helpers.field_funcs import FieldSpec
from gcs_cli.parsing import globus_command, include_private_policies_opt, page_size_opt
from gcs_cli.prettyprint import FORMAT_TEXT_TABLE, formatted_print
from gcs_cli.services.gcs import get_client

STANDARD_FIELDS: FieldSpec = [
    ("ID", "id"),
    ("Display Name", "display_name"),
    ("Owner", default_identity_id_resolver.field),
    ("Collection Type", "collection_type"),
    ("Storage Gateway ID", "storage_gateway_id"),
    ("Created", lambda obj: obj.get("created_at", "Not supported")),
    ("Last Access", lambda obj: obj.get("last_access")),
]


class FilterType(click.ParamType):
    """
    Let people use "mapped-collections" or "mapped_collections" as a nicety.
    Also parse the created_at <=> date and last_access <=> date options
    """

    name = "filter"
    basic_filters = [
        "mapped_collections",
        "guest_collections",
        "managed_by_me",
        "created_by_me",
    ]
    filters = basic_filters + [
        r"created_at\s*([<>]=?|=)\s*\d{4}-\d{2}-\d{2}",
        r"last_access\s*([<>]=?|=)\s*\d{4}-\d{2}-\d{2}",
    ]
    filter_re = re.compile("^(" + "|".join(filters) + ")$")

    @staticmethod
    def hyphen_to_underscore(value: str) -> str:
        def dehyphen(s: re.Match[str]) -> str:
            return s.group(1).replace("-", "_")

        return re.sub("^([a-z_-]*)", dehyphen, value).replace(" ", "")

    def convert(
        self,
        value: str,
        param: t.Optional[click.Parameter],
        ctx: t.Optional[click.Context],
    ) -> t.Optional[str]:
        if value is None:
            return None

        value = self.hyphen_to_underscore(value)
        if not FilterType.filter_re.match(value):
            raise click.BadParameter(
                f"Unable to parse filter {value}", ctx=ctx, param=param
            )

        return t.cast(str, super().convert(value, param, ctx))


@globus_command("list", short_help="List Collections")
@include_private_policies_opt
@requires_login
@click.option(
    "--mapped-collection-id",
    default=None,
    type=click.UUID,
    help=(
        "Filter results to Guest Collections on a specific Mapped Collection. "
        "This is the ID of the Mapped Collection"
    ),
)
@click.option(
    "--storage-gateway-id",
    default=None,
    type=click.UUID,
    help=(
        "Filter results to Collections on a specific Storage Gateway. "
        "This is the ID of the Storage Gateway"
    ),
)
@click.option(
    "--filter",
    "filters",
    multiple=True,
    # choices are shown with hyphens, but the command will receive them with underscores
    type=FilterType(),
    help="""\
Filter results to one of the specified categories of collections. Can be applied
multiple times. Note that mutually exclusive filters are allowed and will find no
results.

The filters are as follows

\b
mapped-collections:
  Only show collections with collection_type="mapped"

\b
guest-collections:
  Only show collections with collection_type="guest"

\b
managed-by-me:
  Only show collections where one of your identities has a role

\b
created-by-me:
  Only show collections where one of your identities was the creator

\b
created-at [< | <= | = | => | >] YYYY-MM-DD:
  Only show collections which were created before, after, or on a certain date.

\b
last-access [< | <= | = | => | >] YYYY-MM-DD:
  Only show collections which have been last used before, after, or on a
  certain date.
""",
)
@page_size_opt
def collection_list(
    include_private_policies: bool,
    mapped_collection_id: uuid.UUID,
    storage_gateway_id: uuid.UUID | None,
    filters: t.Sequence[str],
    page_size: int | None,
) -> None:
    """
    List Collections on the current Endpoint
    """
    client = get_client()
    params: dict[str, str | int] = {}
    if mapped_collection_id:
        params["mapped_collection_id"] = str(mapped_collection_id)
    if storage_gateway_id:
        params["storage_gateway_id"] = str(storage_gateway_id)
    if filters:
        params["filter"] = ",".join(filters)
    if include_private_policies:
        params["include"] = "private_policies"
    if page_size:
        params["page_size"] = page_size

    res: Paginator = client.paginated.get_collection_list(**params)
    res_data = [r for r in res.items()]

    earliest_last_access: date | None = None
    any_unknown = False
    missing_last_access = [
        r
        for r in res_data
        if r.get("created_at") is not None and r.get("last_access") is None
    ]
    if missing_last_access:
        ep = client.get_endpoint()
        ela = ep.get("earliest_last_access")
        if ela:
            earliest_last_access = date.fromisoformat(ela)
    for obj in res_data:
        obj["last_access"] = get_last_access(obj, earliest_last_access)
        if obj["last_access"] == "*":
            any_unknown = True

    formatted_print(res_data, text_format=FORMAT_TEXT_TABLE, fields=STANDARD_FIELDS)
    if any_unknown and earliest_last_access:
        click.echo(f"\n* No recorded access, no info prior to {earliest_last_access}")
