import uuid
from typing import Union

import click

from gcs_cli.helpers import record_print_result_message, requires_login
from gcs_cli.parsing import UUIDorUsername, collection_id_arg, globus_command
from gcs_cli.sdk_client.data import OwnerStringDocument
from gcs_cli.services.auth import get_client as get_auth_client
from gcs_cli.services.gcs import get_client


@globus_command(
    "set-owner-string", short_help="Set the advertised owner string for a collection"
)
@collection_id_arg
@click.argument("OWNER", type=UUIDorUsername())
@requires_login
def collection_set_owner_string(
    collection_id: uuid.UUID, owner: Union[str, uuid.UUID]
) -> None:
    """
    Update the advertised owner string for the given collection.

    This command may only be performed by an collection administrator, and
    can only set the advertised owner to an identity within the caller's identity
    set which has the collection administrator role.
    """
    client = get_client()
    owner_id = None

    if isinstance(owner, uuid.UUID):
        owner_id = str(owner)
    else:
        auth_client = get_auth_client()
        res = auth_client.get_identities(usernames=owner)["identities"]

        if not res:
            res = []
        for identity in res:
            if identity["username"] == owner:
                owner_id = identity["id"]
                break
        if not owner_id:
            click.echo("No such identity {}".format(owner), err=True)
            click.get_current_context().exit(1)

    doc = OwnerStringDocument(identity_id=owner_id)

    res = client.put_collection_owner_string(collection_id, data=doc)
    record_print_result_message(res)
