import uuid
from typing import Any, Optional, Tuple

from gcs_cli.commands.storage_gateway._common import make_update_command
from gcs_cli.parsing import (
    globus_command,
    google_drive_options,
    storage_gateway_create_and_update_params,
    storage_gateway_id_arg,
)
from gcs_cli.sdk_client.data import GoogleDriveStoragePolicies, StorageGatewayDocument
from gcs_cli.services.gcs import GOOGLE_DRIVE_CONNECTOR_ID


@globus_command("google-drive", short_help="Update a Google Drive Storage Gateway")
@storage_gateway_id_arg
@storage_gateway_create_and_update_params(create=False)
@google_drive_options(create=False)
@make_update_command
def google_drive_update(
    storage_gateway_id: uuid.UUID,
    *,
    google_client_id: Optional[str],
    google_client_secret: Optional[str],
    google_drive_user_api_rate_quota: Optional[int],
    google_allow_any_account: Optional[bool],
    **kwargs: Any
) -> Tuple[uuid.UUID, StorageGatewayDocument]:
    """
    Update an existing Google Drive Storage Gateway.
    Any options given will replace existing values in the storage gateway's definition.
    Any options not given will retain their existing values.
    """
    policies = None
    if any(
        x is not None
        for x in (
            google_client_id,
            google_client_secret,
            google_drive_user_api_rate_quota,
            google_allow_any_account,
        )
    ):
        policies = GoogleDriveStoragePolicies(
            client_id=google_client_id,
            secret=google_client_secret,
            user_api_rate_quota=google_drive_user_api_rate_quota,
            allow_any_account=google_allow_any_account,
        )

    return (
        storage_gateway_id,
        StorageGatewayDocument(
            connector_id=GOOGLE_DRIVE_CONNECTOR_ID, policies=policies, **kwargs
        ),
    )
