import logging
import uuid
from typing import Any, Dict, Iterable, Union

from globus_sdk import GlobusHTTPResponse
from globus_sdk.response import IterableResponse
from globus_sdk.services.auth import AuthClient as SDKAuthClient

log = logging.getLogger(__name__)

# Todo: These response definitions exist here because the SDK doesn't yet
# support projects or policies. Once they do, these can be removed.


class GetProjectsResponse(IterableResponse):
    default_iter_key = "projects"


class GetPoliciesResponse(IterableResponse):
    default_iter_key = "policies"


class GetPolicyResponse(GlobusHTTPResponse):
    @property
    def data(self) -> Any:
        return self._parsed_json["policy"]


class GetClientsResponse(IterableResponse):
    default_iter_key = "clients"


class GetClientResponse(GlobusHTTPResponse):
    @property
    def data(self) -> Any:
        return self._parsed_json["client"]


class GetClientCredentialResponse(GlobusHTTPResponse):
    @property
    def data(self) -> Any:
        return self._parsed_json["credential"]


class GetScopesResponse(IterableResponse):
    default_iter_key = "scopes"


class GetScopeResponse(GlobusHTTPResponse):
    @property
    def data(self) -> Any:
        return self._parsed_json["scopes"]


class MissingConsent(Exception):
    def __init__(self, scope: str, cause: Exception):
        self.scope = scope
        self.cause = cause

    def __str__(self) -> str:
        return str(self.cause)


class AuthClient(SDKAuthClient):
    #
    # PROJECTS API
    #
    def get_project_list(self, **params: Any) -> GetProjectsResponse:
        """
        GET /v2/api/projects
        """
        path = "/v2/api/projects"
        return GetProjectsResponse(self.get(path, query_params=params))

    #
    # POLICY API
    #
    def create_auth_policy(
        self, data: Dict[str, Any], **params: Any
    ) -> GetPolicyResponse:
        """
        POST /v2/api/policies

        ** Parameters **

        ``data`` (*dict*)
        Data in the format of a Auth Policy document.
        """
        path = "/v2/api/policies"
        return GetPolicyResponse(
            self.post(path, data=data, query_params=params),
        )

    def get_auth_policy_list(self, **params: Any) -> GetPoliciesResponse:
        """
        GET /v2/api/policies
        """
        path = "/v2/api/policies"
        return GetPoliciesResponse(self.get(path, query_params=params))

    def get_auth_policy(
        self, auth_policy_id: Union[uuid.UUID, str], **params: Any
    ) -> GetPolicyResponse:
        """
        GET /v2/api/policies/{auth_policy_id}
        """
        path = f"/v2/api/policies/{auth_policy_id}"
        return GetPolicyResponse(self.get(path, query_params=params))

    def update_auth_policy(
        self, auth_policy_id: Union[uuid.UUID, str], data: Dict[str, Any], **params: Any
    ) -> GetPolicyResponse:
        """
        PUT /v2/api/policies/{auth_policy_id}

        ** Parameters **

        ``data`` (*dict*)
        Data in the format of a Auth Policy document.
        """
        path = f"/v2/api/policies/{auth_policy_id}"
        return GetPolicyResponse(
            self.put(path, data=data, query_params=params),
        )

    def delete_auth_policy(
        self, auth_policy_id: Union[uuid.UUID, str], **params: Any
    ) -> GetPolicyResponse:
        """
        DELETE /v2/api/policies/{auth_policy_id}

        ** Parameters **

        """
        path = f"/v2/api/policies/{auth_policy_id}"
        return GetPolicyResponse(
            self.delete(path, query_params=params),
        )

    #
    # CLIENT API
    #
    def create_client(self, data: Dict[str, Any], **params: Any) -> GetClientResponse:
        """
        POST /v2/api/clients

        ** Parameters **

        ``data`` (*dict*)
        Data in the format of a client document.
        """
        path = "/v2/api/clients"
        return GetClientResponse(
            self.post(path, data=data, query_params=params),
        )

    def get_client_list(self, **params: Any) -> GetClientsResponse:
        """
        GET /v2/api/clients
        """
        path = "/v2/api/clients"
        return GetClientsResponse(self.get(path, query_params=params))

    def delete_client(
        self, client_id: Union[uuid.UUID, str], **params: Any
    ) -> GetClientResponse:
        """
        DELETE /v2/api/clients/{client_id}
        """
        path = f"/v2/api/clients/{client_id}"
        return GetClientResponse(
            self.delete(path, query_params=params),
        )

    #
    # CREDENTIAL API
    #
    def create_credential(
        self, client_id: Union[uuid.UUID, str], data: Dict[str, Any], **params: Any
    ) -> GetClientCredentialResponse:
        """
        POST /v2/api/clients/<client_id>/credentials

        ** Parameters **

        ``data`` (*dict*)
        Data in the format of a credential document.
        """
        path = f"/v2/api/clients/{client_id}/credentials"
        return GetClientCredentialResponse(
            self.post(path, data=data, query_params=params),
        )

    #
    # SCOPE API
    #
    def create_scope(
        self, client_id: Union[uuid.UUID, str], data: Dict[str, Any], **params: Any
    ) -> GetScopeResponse:
        """
        POST /v2/api/clients/<client_id>/scopes

        ** Parameters **

        ``data`` (*dict*)
        Data in the format of a scope document.
        """
        path = f"/v2/api/clients/{client_id}/scopes"
        return GetScopeResponse(
            self.post(path, data=data, query_params=params),
        )

    def get_scopes(self, scopes: Iterable[str], **params: Any) -> GetScopesResponse:
        """
        GET /v2/api/scopes?scope_strings=<comma-separated-list-of-scope-strings>
        """
        path = "/v2/api/scopes"
        return GetScopesResponse(
            self.get(path, query_params={"scope_strings": ",".join(scopes), **params}),
        )

    def request(self, *args: Any, **kwargs: Any) -> GlobusHTTPResponse:
        projects_scope = "urn:globus:auth:scope:auth.globus.org:manage_projects"
        try:
            return super().request(*args, **kwargs)
        except self.error_class as e:
            if (
                e.http_status == 403
                and e.message is not None
                and projects_scope in e.message
            ):  # noqa: B306
                raise MissingConsent(projects_scope, e)
            raise
