import re
from typing import (
    cast,
    ClassVar,
    Dict,
    List,
    Optional,
    TYPE_CHECKING,
    Type,
)

from .base import (
    Check,
    CheckInvoker,
    MissingPropertyCheck,
    ObjectIsNotNoneCheck,
    PropertyConstraintCheck,
)
from ..parsers.gridftp_options import options as gridftp_options
from ..util.docs import migration_guide_url

if TYPE_CHECKING:
    from ..models.mapped_collection import MappedCollection  # noqa: F401


class MappedCollectionCheck(CheckInvoker):
    checkers: ClassVar[List[Type[Check["MappedCollection"]]]] = []
    object_name = "mapped_collection"


@MappedCollectionCheck.checker
class MC01(ObjectIsNotNoneCheck["MappedCollection"]):
    ...


@MappedCollectionCheck.property_checker(property_name="display_name")
class MC02(MissingPropertyCheck["MappedCollection"]):
    ...


@MappedCollectionCheck.property_checker(property_name="force_verify")
class MC03(MissingPropertyCheck["MappedCollection"]):
    ...


@MappedCollectionCheck.property_checker(property_name="disable_verify")
class MC04(MissingPropertyCheck["MappedCollection"]):
    ...


@MappedCollectionCheck.property_checker(property_name="enable_https")
class MC05(MissingPropertyCheck["MappedCollection"]):
    ...


@MappedCollectionCheck.property_checker(property_name="allow_guest_collections")
class MC06(MissingPropertyCheck["MappedCollection"]):
    ...


@MappedCollectionCheck.property_checker(property_name="sharing_users_allow")
class MC07(PropertyConstraintCheck["MappedCollection"]):
    constraint_name = "sharing disabled but sharing_users_allow is set"

    def value_is_ok(self, obj: Optional["MappedCollection"]) -> bool:
        ok = True

        if obj and obj.allow_guest_collections is False and obj.sharing_users_allow:
            ok = False

        return ok


@MappedCollectionCheck.property_checker(property_name="sharing_users_deny")
class MC08(PropertyConstraintCheck["MappedCollection"]):
    constraint_name = "sharing disabled but sharing_users_deny is set"

    def value_is_ok(self, obj: Optional["MappedCollection"]) -> bool:
        ok = True

        if obj and obj.allow_guest_collections is False and obj.sharing_users_deny:
            ok = False

        return ok


@MappedCollectionCheck.property_checker(property_name="sharing_groups_allow")
class MC09(PropertyConstraintCheck["MappedCollection"]):
    constraint_name = "sharing disabled but sharing_groups_allow is set"

    def value_is_ok(self, obj: Optional["MappedCollection"]) -> bool:
        ok = True

        if obj and obj.allow_guest_collections is False and obj.sharing_groups_allow:
            ok = False

        return ok


@MappedCollectionCheck.property_checker(property_name="sharing_groups_deny")
class MC10(PropertyConstraintCheck["MappedCollection"]):
    constraint_name = "sharing disabled but sharing_groups_deny is set"

    def value_is_ok(self, obj: Optional["MappedCollection"]) -> bool:
        ok = True

        if obj and obj.allow_guest_collections is False and obj.sharing_groups_deny:
            ok = False

        return ok


@MappedCollectionCheck.property_checker(property_name="gridftp_config_data")
class MC11(PropertyConstraintCheck["MappedCollection"]):
    constraint_name = "Custom GridFTP sharing client"

    def __init__(self) -> None:
        super().__init__()
        self.sharing_dn: Optional[str] = None

    def value_is_ok(self, obj: Optional["MappedCollection"]) -> bool:
        ok = True

        if obj and obj.gridftp_config_data:
            sharing_dn = obj.gridftp_config_data.get("sharing_dn")
            if (
                sharing_dn is not None
                and sharing_dn
                != "/C=US/O=Globus Consortium/OU=Globus Online/OU=Transfer User/CN=__transfer__"
            ):
                self.sharing_dn = cast(str, sharing_dn)
                ok = False
        return ok

    @property
    def diagnostic(self) -> Optional[str]:
        if self.sharing_dn:
            return "Custom sharing client dn not supported in GCSv5"
        return None


@MappedCollectionCheck.property_checker(property_name="gridftp_config_data")
class MC12(PropertyConstraintCheck["MappedCollection"]):
    constraint_name = "Unsupported connector"
    supported_connectors = ["file", "s3", "blackpearl", "ceph", "hpss_local"]

    def __init__(self) -> None:
        super().__init__()
        self.connector: Optional[str] = None

    def value_is_ok(self, obj: Optional["MappedCollection"]) -> bool:
        ok = True

        if obj and obj.gridftp_config_data:
            connector = obj.gridftp_config_data.get("load_dsi_module")
            if connector is not None and connector not in self.supported_connectors:
                self.connector = cast(str, connector)
                ok = False
        return ok

    @property
    def diagnostic(self) -> Optional[str]:
        if self.connector:
            return f"The following connectors are supported by this tool: {self.supported_connectors}"
        return None


@MappedCollectionCheck.property_checker(property_name="gridftp_config_data")
class MC13(PropertyConstraintCheck["MappedCollection"]):
    """
    Check for unsupported configurations which have alternatives in GCSv5. If
    found, provide a warning that this functionality has changed in GCSv5, but
    there is a similar feature that can be set after migration is complete.
    """

    constraint_name = (
        "The GridFTP configuration uses features that have changed in GCSv5, "
        f"see {migration_guide_url}#MC13 "
        "for information on how to configure the feature in GCSv5."
    )
    unsupported_features: Dict[str, str] = {
        "banner": "Messages displayed at GridFTP Login",
        "network": "Network addresses and port ranges",
    }

    unsupported_configs: Dict[str, str] = {
        "banner": "banner",
        "banner_file": "banner",
        "banner_terse": "banner",
        "login_msg": "banner",
        "login_msg_file": "banner",
        "control_interface": "network",
        "data_interface": "network",
        "hostname": "network",
    }

    # For the banner option, the GCSv4 setup doesn't set it explicitly, but the
    # GridFTP parser generates it based on hostname, gridftp version, and
    # version_tag. If it looks like the default form, we'll ignore it in this
    # check
    ignore_banner_pattern = (
        r"[a-zA-Z0-9.-]+ GridFTP Server [0-9.]+ "
        r"\([a-z]+[0-9]+, [0-9-]+\) \[Globus Toolkit [a-zA-Z0-9. -]*\] ready."
    )

    def __init__(self) -> None:
        super().__init__()

    def value_is_ok(self, obj: Optional["MappedCollection"]) -> bool:
        ok = True
        bad_config: Dict[str, List[str]] = {}

        if obj and obj.gridftp_config_data:
            for option in self.unsupported_configs:
                val = obj.gridftp_config_data[option]
                if (
                    option == "banner"
                    and isinstance(val, str)
                    and re.match(self.ignore_banner_pattern, val)
                ):
                    continue
                elif obj.gridftp_config_data[option] != gridftp_options[option].default:
                    feature = self.unsupported_configs[option]
                    if feature not in bad_config:
                        bad_config[feature] = []
                    bad_config[feature].append(option)
            for option in ["port_range", "$GLOBUS_TCP_SOURCE_RANGE"]:
                if (
                    obj.gridftp_config_data[option] != gridftp_options[option].default
                    and obj.gridftp_config_data[option] != "50000,51000"
                ):
                    feature = "network"
                    if feature not in bad_config:
                        bad_config[feature] = []
                    bad_config[feature].append(option)

        if bad_config:
            print(
                "\n".join(
                    [f"{self.name}:", self.constraint_name]
                    + [
                        f"    {self.unsupported_features[feature]}: {', '.join(feature_configs)}"
                        for feature, feature_configs in bad_config.items()
                    ]
                )
            )
        return ok


@MappedCollectionCheck.property_checker(property_name="gridftp_config_data")
class MC14(PropertyConstraintCheck["MappedCollection"]):
    """
    Check for unsupported configurations which have no alternatives in GCSv5, but
    can be manually managed if needed.
    """

    constraint_name = (
        "The GridFTP configuration uses features that have changed in GCSv5, "
        f"see {migration_guide_url}#MC14 "
        "for information on how to configure the alternatives."
    )
    unsupported_features: Dict[str, str] = {
        "xnetmgr": "Network manager",
        "logging": "Manual logging configuration",
        "debugging": "Debugging configuration",
        "usagestats": "Custom usage stats target",
        "data_tuning": "Custom data access tuning",
        "udt": "UDT protocol support",
        "root": "Allowing users to be mapped to the root user",
        "disabled": "Allowing users with disabled accounts to log in",
        "network_tuning": "Tuning network connections",
        "process": "Custom process environment and user",
    }

    unsupported_configs: Dict[str, str] = {
        "xnetmgr": "xnetmgr",
        "log_module": "logging",
        "log_unique": "logging",
        "log_transfer": "logging",
        "log_filemode": "logging",
        "exec": "debugging",
        "fork": "debugging",
        "fork_fallback": "debugging",
        "single": "debugging",
        "debug": "debugging",
        "ignore_bad_threads": "debugging",
        "bad_signal_exit": "debugging",
        "blocksize": "data_tuning",
        "sync_writes": "data_tuning",
        "direct": "data_tuning",
        "perms": "data_tuning",
        "file_timeout": "data_tuning",
        "slow_dirlist": "data_tuning",
        "checksum_throttle": "data_tuning",
        "allow_udt": "udt",
        "chdir": "debugging",
        "chdir_to": "debugging",
        "allow_root": "root",
        "allow_disabled_login": "disabled",
        "connections_max": "network_tuning",
        "control_preauth_timeout": "network_tuning",
        "control_idle_timeout": "network_tuning",
        "connections_disabled": "network_tuning",
        "offline_msg": "network_tuning",
        "threads": "process",
    }
    dsi_ignore = {
        "blackpearl": ["threads", "process_user"],
        "s3": ["threads", "process_user"],
    }

    def __init__(self) -> None:
        super().__init__()

    def value_is_ok(self, obj: Optional["MappedCollection"]) -> bool:
        ok = True
        bad_config: Dict[str, List[str]] = {}

        if obj and obj.gridftp_config_data:
            connector = cast(str, obj.gridftp_config_data["load_dsi_module"])

            ignore = self.dsi_ignore.get(connector, [])

            for option in self.unsupported_configs:
                if option in ignore:
                    continue
                if obj.gridftp_config_data[option] != gridftp_options[option].default:
                    feature = self.unsupported_configs[option]
                    if feature not in bad_config:
                        bad_config[feature] = []
                    bad_config[feature].append(option)
            for option in obj.gridftp_config_data.keys():
                if re.match(r"\$.*_DEBUG", option):
                    feature = "debugging"
                    if feature not in bad_config:
                        bad_config[feature] = []
                    bad_config[feature].append(option)

            usage_stats_target = obj.gridftp_config_data["usage_stats_target"]
            if (
                usage_stats_target is not None
                and usage_stats_target != "usage-stats.globus.org"
            ):
                feature = "usagestats"
                if feature not in bad_config:
                    bad_config[feature] = []
                bad_config[feature].append("usage_stats_target")
            option = "log_level"
            if (
                obj.gridftp_config_data[option] != gridftp_options[option].default
                and obj.gridftp_config_data[option] != "ERROR,WARN"
            ):
                feature = "logging"
                if feature not in bad_config:
                    bad_config[feature] = []
                bad_config[feature].append(option)
            option = "log_single"
            if (
                obj.gridftp_config_data[option] != gridftp_options[option].default
                and obj.gridftp_config_data[option] != "/var/log/gridftp.log"
            ):
                feature = "logging"
                if feature not in bad_config:
                    bad_config[feature] = []
                bad_config[feature].append(option)
            option = "port"
            if (
                obj.gridftp_config_data[option] != gridftp_options[option].default
                and obj.gridftp_config_data[option] != 2811
            ):
                feature = "network_tuning"
                if feature not in bad_config:
                    bad_config[feature] = []
                bad_config[feature].append(option)

        if bad_config:
            print(
                "\n".join(
                    [f"{self.name}:", self.constraint_name]
                    + [
                        f"    {self.unsupported_features[feature]}: {', '.join(feature_configs)}"
                        for feature, feature_configs in bad_config.items()
                    ]
                )
            )
        return ok


@MappedCollectionCheck.property_checker(property_name="gridftp_config_data")
class MC15(PropertyConstraintCheck["MappedCollection"]):
    """
    Check for unsupported configurations which have no alternatives in GCSv5, and
    are incompatible with GCSv5
    """

    constraint_name = (
        "The GridFTP configuration uses features are incompatible with GCSv5, "
        f"see {migration_guide_url}#MC15 "
        "for information."
    )
    unsupported_features: Dict[str, str] = {
        "ssh": "SSH FTP",
        "chroot": "Running the GridFTP server in a chroot",
        "striped_server": "Configuring the GridFTP server as a striped server",
        "anon": "Allowing anonymous FTP access to the GridFTP server",
        "password": "Password authentication to the GridFTP server",
        "sharing_control": "Control-channel management of sharing policies and commands",
        "stacks": "Custom network and driver stacks",
        "authz": "Custom authorization modules",
        "process": "Custom process environment and user",
        "control": "Custom allow/deny IP ranges for control connections",
        "symlink": "Allowing symlinks to leave the path restrictions for a collection",
        "globus_location": "Setting a non-default globus location",
    }

    unsupported_configs: Dict[str, str] = {
        "ssh": "ssh",
        "chroot_path": "chroot",
        "ipc_interface": "striped_server",
        "ipc_allow_from": "striped_server",
        "ipc_deny_from": "striped_server",
        "secure_ipc": "striped_server",
        "ipc_auth_mode": "striped_server",
        "ipc_user_name": "striped_server",
        "ipc_subject": "striped_server",
        "ipc_credential": "striped_server",
        "ipc_cookie": "striped_server",
        "ipc_port": "striped_server",
        "ipc_idle_timeout": "striped_server",
        "ipc_connect_timeout": "striped_server",
        "remote_nodes": "striped_server",
        "hybrid": "striped_server",
        "data_node": "striped_server",
        "stripe_blocksize": "striped_server",
        "stripe_count": "striped_server",
        "brain": "striped_server",
        "stripe_layout": "striped_server",
        "stripe_blocksize_locked": "striped_server",
        "stripe_layout_locked": "striped_server",
        "stripe_mode": "striped_server",
        "allow_anonymous": "anon",
        "anonymous_names_allowed": "anon",
        "anonymous_user": "anon",
        "anonymous_group": "anon",
        "pw_file": "password",
        "sharing_control": "sharing_control",
        "allowed_modules": "stacks",
        "dc_whitelist": "stacks",
        "fs_whitelist": "stacks",
        "popen_whitelist": "stacks",
        "dc_default": "stacks",
        "fs_default": "stacks",
        "cas": "authz",
        "acl": "authz",
        "auth_level": "process",
        "process_user": "process",
        "process_group": "process",
        "inetd": "process",
        "daemon": "process",
        "detach": "process",
        "pidfile": "process",
        "use_home_dirs": "process",
        "home_dir": "process",
        "rp_follow_symlinks": "symlink",
        "disable_command_list": "sharing_control",
        "globus_location": "globus_location",
    }

    def __init__(self) -> None:
        super().__init__()

    def value_is_ok(self, obj: Optional["MappedCollection"]) -> bool:
        ok = True
        bad_config: Dict[str, List[str]] = {}

        if obj and obj.gridftp_config_data:
            for option in self.unsupported_configs:
                if obj.gridftp_config_data[option] != gridftp_options[option].default:
                    feature = self.unsupported_configs[option]
                    if feature not in bad_config:
                        bad_config[feature] = []
                    bad_config[feature].append(option)
            option = "allow_from"
            val = obj.gridftp_config_data[option]
            if val is not None and val != "54.237.254.192/29":
                feature = "control"
                if feature not in bad_config:
                    bad_config[feature] = []
                bad_config[feature].append(option)

            option = "deny_from"
            val = obj.gridftp_config_data[option]
            if val is not None:
                feature = "control"
                if feature not in bad_config:
                    bad_config[feature] = []
                bad_config[feature].append(option)

        if bad_config:
            print(
                "\n".join(
                    [self.constraint_name]
                    + [
                        f"    {self.unsupported_features[feature]}: {', '.join(feature_configs)}"
                        for feature, feature_configs in bad_config.items()
                    ]
                )
            )
        return ok


@MappedCollectionCheck.property_checker(property_name="gridftp_config_data")
class MC16(PropertyConstraintCheck["MappedCollection"]):
    """
    Check for process_user in blackpearl connector. If set, emit a non-fatal
    diagnostic that the blackpearl user map file must be readable by the
    "gcsweb" user.
    """

    constraint_name = (
        "The process_user for a GCSv5 endpoint is always set to the gcsweb "
        "user, but your v4 endpoint uses a different value."
    )

    def __init__(self) -> None:
        super().__init__()

    def value_is_ok(self, obj: Optional["MappedCollection"]) -> bool:
        ok = True

        if obj and obj.gridftp_config_data:
            connector = obj.gridftp_config_data["load_dsi_module"]
            if connector != "blackpearl":
                return ok

            process_user = obj.gridftp_config_data["process_user"]
            if (
                process_user != gridftp_options["process_user"].default
                and process_user != "gcsweb"
            ):
                print(
                    "Your GCSv4 configuration uses the process_user value of "
                    f"{process_user}. This is the name of the user which owns the "
                    "blackpearl user map file. In GCSv5, this file must be readable "
                    "by the gcsweb user."
                )
        return ok
