from __future__ import annotations

import ctypes
import ctypes.util
import os
import typing as t

from cachelib import SimpleCache
from flask import current_app

import globus.manager.api.models as m
import globus.manager.api.schemas.connectors.hpss  # noqa

if t.TYPE_CHECKING:
    from collections import abc

    from globus.manager.api.app import GCSManagerApp

# The connector classes are loaded as singletons during app initialization. If
# we use `gunicorn --preload`, the init code will be executed once in the main
# gunicorn process and the results will be available as copy-on-write memory to
# all forked processes. Without `--preload`, the init code is executed once per
# worker process. This creates a couple interesting challenges:
#
# - Each worker thread will access its process's singleton instance concurrently,
#   so care must be taken to remain thread-safe. We could cheat when thread count
#   is 1, but that would make things fragile should the thread count change.
# - There is no request context during initialization.
# - Some initialization may not have completed.
# - Updates to any loaded libraries will require restarting GCSM
#
# Remember that most sites will not have the HPSS DSI installed, so logging errors
# on startup is probably undesirable.


class _NoHomeDir:
    """
    Sentinel value used to disambiguate return values of SimpleCache.get() when
    dealing with race conditions where the key expires between SimpleCache.has()
    and SimpleCache.get().
    """

    def __bool__(self) -> bool:
        return False

    def __str__(self) -> str:
        return "None"


NO_HOME_DIR = _NoHomeDir()


class HPSSConnector:
    CONNECTOR_ID = "fb656a17-0f69-4e59-95ff-d0a62ca7bdf5"
    DISPLAY_NAME = "HPSS"
    model = m.HPSSStorageGateway

    _get_home_directory_func: abc.Callable[
        [bytes, bytes, bytes, abc.Callable[[bytes], None]], int
    ] | None = None
    _get_home_directory_func_ex: abc.Callable[
        [bytes, bytes, bytes, bytes, abc.Callable[[bytes], None]], int
    ] | None = None

    def __init__(self, app: GCSManagerApp) -> None:
        """
        __init__ is called during app initialization; there is no request context!
        """
        self._load_get_home_directory_funcs()
        self._cache = SimpleCache(threshold=100, default_timeout=600)

    def _load_get_home_directory_funcs(self) -> None:
        """
        Load the home directory functions from the DSI. It takes one of two forms:

        int
        get_home_directory(char *  AuthenticationMech, // Auth type krb5,unix
                           char *  Authenticator, // Location of credentials <type>:<path>
                           char *  UserName, // User to switch to
                           void (*callback)(char *))
        int
        get_home_directory_ex(char *  AuthenticationMech, // Auth type krb5,unix
                              char *  Authenticator, // Location of credentials <type>:<path>
                              char *  LoginName, // 'hpssftp'
                              char *  UserName, // User to switch to
                              void (*callback)(char *))

        callback(str) passes us a home directory or error message, depending on
        the return value of get_home_directory()
        """
        gridftp_dll = self._load_library("globus_gridftp_server")
        if not gridftp_dll:
            return None

        hpss_dsi_dll = self._load_library("globus_gridftp_server_hpss")
        if not hpss_dsi_dll:
            return None

        #
        # Find the newer function first.
        #

        func = getattr(hpss_dsi_dll, "get_home_directory_ex", None)

        if func:
            func.restype = ctypes.c_int
            func.argtypes = [
                ctypes.c_char_p,  # AuthenticationMech
                ctypes.c_char_p,  # Authenticator
                ctypes.c_char_p,  # LoginName
                ctypes.c_char_p,  # UserName
                ctypes.CFUNCTYPE(None, ctypes.c_char_p),
            ]
            self._get_home_directory_func_ex = func
            return

        #
        # Fallback on the older function.
        #

        func = getattr(hpss_dsi_dll, "get_home_directory", None)
        if not func:
            # Log this with severity error since if we got this far, the DSI is installed
            # and it should have the home directory function.
            current_app.logger.error(
                "HPSS storage gateway failed to load the helper function. Install "
                "the latest version of the HPSS DSI and restart the GCS manager."
            )
            return None

        func.restype = ctypes.c_int
        func.argtypes = [
            ctypes.c_char_p,  # AuthenticationMech
            ctypes.c_char_p,  # Authenticator
            ctypes.c_char_p,  # UserName
            ctypes.CFUNCTYPE(None, ctypes.c_char_p),
        ]
        self._get_home_directory_func = func

    def _load_library(self, library: str) -> ctypes.CDLL | None:
        """
        Wrap library loading with useful error logging.

        Python ctypes does not explicitly support RTLD_LAZY but here's a hack to
        passing it through to Linux if use another flag (ex. RTLD_GLOBAL).
        """

        lib_name = ctypes.util.find_library(library)
        if not lib_name:
            # Log this error as debug to prevent non-HPSS sites from seeing it
            current_app.logger.debug(f"HPSS storage gateway cannot find lib{library}")
            return None

        try:
            # This will throw OSError if any of the library's dependencies are missing
            return ctypes.CDLL(lib_name, mode=os.RTLD_LAZY | ctypes.RTLD_GLOBAL)
        except OSError as e:
            current_app.logger.error(
                f"HPSS storage gateway failed to load {library}: {e}"
            )
            return None

    def home_directory(
        self,
        authentication_mech: str,
        authenticator: str,
        login_name: str,
        username: str,
    ) -> str | None:
        """
        Lookup username's home directory and cache the response. If the lookup
        is successful, cache the value for 10 minutes. If the lookup fails, cache
        the value for 10 seconds.
        """
        current_app.logger.debug(f"Looking up home directory for {username}")

        home_dir: str | _NoHomeDir | None = self._cache.get(username)
        if home_dir is not None:
            current_app.logger.debug(f"Using cached home directory {home_dir}")
            return home_dir if isinstance(home_dir, str) else None

        home_dir = self._home_directory(
            authentication_mech, authenticator, login_name, username
        )
        self._cache.set(
            username,
            home_dir if isinstance(home_dir, str) else NO_HOME_DIR,
            timeout=600 if isinstance(home_dir, str) else 10,
        )
        current_app.logger.debug(f"User's home directory is {home_dir}")
        return home_dir

    def _home_directory(
        self,
        authentication_mech: str,
        authenticator: str,
        login_name: str,
        username: str,
    ) -> str | None:
        """
        Look up the user's home directory by calling a helper function built into
        the HPSS DSI. Used for expansion of '~' and  $HOME in path restrictions.

        This is called from a request context. Locking is handled in the DSI.
        """

        # This callback is called from the DSI with our string value.
        # On success, it returns 0 and sets 'value' to the user's home directory.
        # Otherwise, returns non-zero and sets 'value' to an error message.
        home_dir: str | None = None

        def callback(value: bytes) -> None:
            nonlocal home_dir
            if value:
                home_dir = value.decode()

        rc: int = 0
        if self._get_home_directory_func_ex:
            # Finally, make the call to get the home directory
            rc = self._get_home_directory_func_ex(
                authentication_mech.encode("utf-8"),
                authenticator.encode("utf-8"),
                login_name.encode("utf-8"),
                username.encode("utf-8"),
                ctypes.CFUNCTYPE(None, ctypes.c_char_p)(callback),
            )
        elif self._get_home_directory_func:
            # Finally, make the call to get the home directory
            rc = self._get_home_directory_func(
                authentication_mech.encode("utf-8"),
                authenticator.encode("utf-8"),
                username.encode("utf-8"),
                ctypes.CFUNCTYPE(None, ctypes.c_char_p)(callback),
            )
        else:
            current_app.logger.error(
                "The HPSS Connector is unable to load the HPSS DSI for home directory lookups. "
                "Make sure the latest version of the HPSS DSI is installed and restart the GCS "
                "manager."
            )
            return None

        # Interpret the output we received in the callback
        if rc:
            current_app.logger.error(
                f"Could not find home directory for {username}. {home_dir}. "
                f"Error code is {rc}."
            )
            return None

        return home_dir
