"""
This module implements a mapping algorithm to
transform an Identity resource from Globus Auth to a user name which is
meaningful in the context of the Connector that is associated with the
Storage Gateway mapping is being peformed on.

The input mapping_data consists of a list of mappings, which are passed in
as a dicts containing the following fields:

*   source: A source expression that contains a string which may include
    interpolations from properties of the Identity resource being
    evaluated.
*   match: A match expression to compare against the source. This
    expression supports a limited regular expression syntax which is
    applied to the source string.
*   output: An output expression which can contain references to properties
    of the input Identity resource or references to parenthesized
    subexpressions from applying the match expression to the source.
*   ignore_case: A boolean flag controlling whether to perform a case
    insensitive match. The default is False.
*   literal: A boolean flag whether to interpet the match expression
    literally, or to use the regular expression language defined below when
    determining whether a match succeeds. If this is set to True,
    references to match subexpressions may not be used in the output field.

Match Expression Language
-------------------------

When the literal flag is not present, the match value is interpreted using
a reduced version of the POSIX extended regular expression, including
special processing for the following tokens:

* :code:`.` any single character match
* :code:`?` for zero or one of the preceding match
* :code:`*` for zero or more of the preceding match
* :code:`|` for alternative match branches
* :code:`()` to group a match and created a numbered submatch for output
  interpolation
* `\\` to escape any of the above, `\\`, or :code:`"` from special
  processing

Any other string in the match expression is matched literally

Identity Data Interpolation
---------------------------

The source and output strings may include values that are interpolated from
a Globus auth identity resource, as described in the
`Globus Auth documentation
<https://docs.globus.org/api/auth/reference/#_identity_resource>`_.

Interpolations are indicated by surrounding the resource property with curly
braces. For example, to interpolate the username property of the identity
resource in the source string, one uses the value :code:`{username}`; likewise,
to interpolate the id property, one uses the string :code:`{id}`. Multiple
values may be interpolated into either the source or output strings.

Subexpression Interpolation
---------------------------

The output string may also contain references to subexpression matches which
are also interpolated into the output. The matching is down by using a similar
syntax the Identity Data Interpolation above, but using integer values to
reference matched subexpression, with indices starting at 0. For example, using
a match expression :code:`(.*)@example\\.org`, one can interpolate the part
before the :code:`@` symbol my using the string :code:`{0}` in the output
expression.

Examples
--------

To match username against a given domain, and return the username as the
output expression:

.. code-block:: json

    {
        "source": "{username}",
        "match": "(.*)@example\\.org",
        "output": "{0}"
    }

To match an identity ID and map to a particular user account

.. code-block:: json

    {
        "source": "{id}",
        "match": "b34bb069-3b68-4d38-b4e6-bc1e4fc95332",
        "output": "application_account"
    }

To rewrite an identity from one domain to another:

.. code-block: json

    {
        "source": "{username}",
        "match": "(.*)@example\\.edu"
        "output": "{0}@data.example.edu"
    }

To map a set of identities to a common account:

.. code-block: json

    {
        "source": "{username}",
        "match": "(alice|bob)@example\\.org",
        "output": "globus"
    }
"""

from __future__ import annotations

import re
from typing import Any, Dict, List, Optional, Pattern

from typing_extensions import TypedDict


class ExpressionIdentityMappingDict(TypedDict, total=False):
    source: str
    match: str
    output: str
    literal: bool
    ignore_case: bool


class CompiledExpressionIdentityMappingDict(TypedDict, total=False):
    source: str
    match: str
    output: str
    literal: bool
    ignore_case: bool
    compiled_match: Pattern


class ExpressionIdentityMapping:
    """
    The ExpressionIdentityMapping class implements a mapping algorithm to
    transform an Identity resource from Globus Auth to a user name which is
    meaningful in the context of the Connector that is associated with the
    Storage Gateway mapping is being performed on.
    """

    DATA_TYPE = "expression_identity_mapping#1.0.0"

    def __init__(
        self,
        mapping_data: List[ExpressionIdentityMappingDict],
        storage_gateway: str | None = None,
        connector: str | None = None,
    ) -> None:
        """
        :param List[ExpressionIdentityMappingDict] mapping_data:
            Identity mapping definitions as described in the API documentation.
        :param str storage_gateway:
            The storage gateway that the mapping will be performed for. This is
            not used in this module
        :param str connector:
            The connector uuid that the mapping will be performed for. This is
            not used in this module
        """
        self.mappings: List[CompiledExpressionIdentityMappingDict] = []
        self.connector = connector
        self.storage_gateway = storage_gateway

        for m in mapping_data:
            mapping: CompiledExpressionIdentityMappingDict = {}
            mapping["source"] = m["source"]
            mapping["compiled_match"] = self._compile_match(
                m["match"], m.get("ignore_case", False), m.get("literal", False)
            )
            mapping["match"] = m["match"]
            mapping["output"] = m["output"]
            if "ignore_case" in m:
                mapping["ignore_case"] = m["ignore_case"]
            if "literal" in m:
                mapping["literal"] = m["literal"]
            self.mappings.append(mapping)

    @staticmethod
    def from_json(
        mapping_document: Dict[str, Any],
        connector: str | None = None,
        storage_gateway: str | None = None,
    ) -> "ExpressionIdentityMapping":
        if mapping_document["DATA_TYPE"] != ExpressionIdentityMapping.DATA_TYPE:
            raise ValueError("Unsupported DATA_TYPE")

        return ExpressionIdentityMapping(
            mapping_document["mappings"],
            connector=connector,
            storage_gateway=storage_gateway,
        )

    @staticmethod
    def _compile_match(
        match: str, ignore_case: bool = False, literal: bool = False
    ) -> Pattern:
        original_match = match
        if not literal:
            bad_escape = re.search(r"(\\[^.?*|()\\0-9])", match)
            if bad_escape:
                raise ValueError(
                    "{}: Invalid \\ expression {}".format(match, bad_escape.group())
                )
            unsupported_re = re.search(r"([*?]\?)", match)
            if unsupported_re:
                raise ValueError(
                    "{}: Invalid ? expression {}".format(match, unsupported_re.group())
                )
            leading_wildcard_re = re.search(r"^([*?])", match)
            if leading_wildcard_re:
                raise ValueError(
                    "{}: Invalid leading wildcard expression {}".format(
                        match, leading_wildcard_re.group()
                    )
                )

            # Escape python regex specials that we don't support
            match = re.sub(r"(?<!\\)([\^\$\+\{\}\[\]])", r"\\\1", match)
        else:
            match = re.sub(r"(?<!\\)([\^\?\*\|\(\)\\\$\+\.\{\}\[\]])", r"\\\1", match)
            match = re.sub(r"(\\[0-9])", r"\\\1", match)

        try:
            return re.compile("^" + match + "$", re.I if ignore_case else 0)
        except Exception as e:
            raise ValueError("Invalid mapping {}: {}", original_match, str(e))

    @staticmethod
    def _interpolate(
        string: str, identity_data: Dict[str, str], matches: List | None = None
    ) -> str:
        if not matches:
            matches = []
        interpolated = string

        for m in re.finditer(r"(?<!\\)(\\\\)*({([^}]*)})", string):
            subexp, subname = m.groups()[1:]
            try:
                matchref = int(subname)
                if matchref >= len(matches):
                    raise ValueError("Invalid reference to {}".format(matchref))
                repl = matches[matchref]
            except ValueError:
                repl = identity_data.get(subname, None)
            if repl is None:
                raise ValueError("Missing identity property {}".format(subname))
            interpolated = interpolated.replace(subexp, repl)
        return interpolated

    def map_identity(self, identity_data: Dict[str, str]) -> Optional[str]:
        """
        Map an identity resource to a connector-specific user name if possible.
        Applies the mapping to identity_data, and if successful returns the
        interpolated output string of the first match. If unsuccessful, returns
        None.
        """
        for m in self.mappings:
            source = self._interpolate(m["source"], identity_data)
            mr = m["compiled_match"].match(source)
            if mr:
                return self._interpolate(m["output"], identity_data, list(mr.groups()))
        return None

    def map_identities(
        self, identity_data: List[Dict[str, str]]
    ) -> List[Dict[str, List[str]]]:
        """
        Map a list of identity resource to a connector-specific user name if
        possible.  Applies the each mapping to each identity resource provided
        in the identity_data list, and if successful
        returns the
        interpolated output string of the first match. If unsuccessful, returns
        None.


        Returns a list of results in the order that the mappings are defined.
        For each list item, the value is a dict mapping the id of an identity
        to the output of the mapping, if the match succeeded.
        """
        results = []
        for m in self.mappings:
            mapping_results: Dict[str, List[str]] = {}
            for identity_datum in identity_data:
                if identity_datum["status"] not in ["used", "private"]:
                    continue

                source = self._interpolate(m["source"], identity_datum)
                mr = m["compiled_match"].match(source)
                if mr:
                    identity_id = identity_datum["id"]
                    if identity_id not in mapping_results:
                        mapping_results[identity_id] = []
                    mapping_results[identity_id].append(
                        self._interpolate(
                            m["output"], identity_datum, list(mr.groups())
                        )
                    )
            results.append(mapping_results)
        return results
