from __future__ import annotations

import binascii
import datetime
import os
import time
import typing as t

import jwt
import requests
from cachelib import SimpleCache
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from flask import current_app

from globus.manager.api.errors import GatewayTimeout
from globus.manager.constants import ConfigurationObjectTag

from .connector import Connector
from .decorators import checks_policy_property
from .storage_gateway import StorageGateway

# cache enterprise access tokens by the EnterpriseID they are for
cache = SimpleCache()
user_info_cache = SimpleCache()
user_token_cache = SimpleCache()


class BoxConnectionError(GatewayTimeout):
    error_description = (
        "Unable to connect to Box. This storage gateway's "
        "storage policies may be misconfigured."
    )


class ClockSkewError(BoxConnectionError):
    error_description = (
        "Unable to connect to Box. Please ensure your system time is correct."
    )


class BoxStorageGateway(StorageGateway):
    CONNECTOR_ID = "7c100eae-40fe-11e9-95a3-9cb6d0d9fd63"
    POLICY_DATA_TYPE = "box_storage_policies#1.2.0"
    CONNECTOR_POLICIES = {
        "box_storage_policies#1.0.0": [
            "boxAppSettings",
            "enterpriseID",
        ]
    }
    CONNECTOR_POLICIES["box_storage_policies#1.1.0"] = CONNECTOR_POLICIES[
        "box_storage_policies#1.0.0"
    ] + [
        "client_id",
        "secret",
        "user_api_rate_limit",
    ]
    CONNECTOR_POLICIES["box_storage_policies#1.2.0"] = CONNECTOR_POLICIES[
        "box_storage_policies#1.1.0"
    ] + ["allow_any_account"]

    POLICY_DATA_TYPES = list(CONNECTOR_POLICIES.keys())
    connector_policy_names = CONNECTOR_POLICIES[POLICY_DATA_TYPE]

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.STORAGE_GATEWAY.name + CONNECTOR_ID
        )
    }

    connector = Connector(
        id=CONNECTOR_ID, display_name="Box", can_post_credential=False
    )

    boxAppSettings = StorageGateway.connector_property("boxAppSettings")

    @checks_policy_property(prop="boxAppSettings")
    def check_boxAppSettings(self, value: dict[str, t.Any]):
        if not isinstance(value, dict):
            raise ValueError()
        if getattr(self, "client_id", None) or getattr(self, "secret", None):
            raise ValueError(
                "Unsupported boxAppSettings property for OAuth configuration"
            )

        required_keys = ["clientID", "clientSecret", "appAuth"]
        for missing in required_keys:
            if missing not in value:
                raise ValueError(
                    "Invalid value for the boxAppSettings property: missing {}.".format(
                        missing
                    )
                )
        required_appauth = ["publicKeyID", "privateKey", "passphrase"]
        for missing in required_appauth:
            if missing not in value["appAuth"]:
                raise ValueError(
                    "Invalid value for the boxAppSettings property: missing {}".format(
                        missing
                    )
                )

    enterpriseID = StorageGateway.connector_property("enterpriseID")

    @checks_policy_property(prop="enterpriseID")
    def check_enterpriseID(self, value: str):
        if not isinstance(value, str):
            raise ValueError()
        if getattr(self, "client_id", None) or getattr(self, "secret", None):
            raise ValueError(
                "Unsupported enterpriseID property for OAuth configuration"
            )

    client_id = StorageGateway.connector_property("client_id")

    @checks_policy_property(prop="client_id")
    def check_client_id(self, value: str):
        if not isinstance(value, str):
            raise ValueError("Invalid client_id")
        if getattr(self, "enterpriseID", None) or getattr(self, "boxAppSettings", None):
            raise ValueError(
                "Unsupported client_id property for enterprise configuration"
            )

    secret = StorageGateway.connector_property("secret")

    @checks_policy_property(prop="secret")
    def checks_secret(self, value: str):
        if not isinstance(value, str):
            raise ValueError("Invalid secret")
        if getattr(self, "enterpriseID", None) or getattr(self, "boxAppSettings", None):
            raise ValueError("Unsupported secret property for enterprise configuration")

    user_api_rate_limit = StorageGateway.connector_property("user_api_rate_limit")

    @checks_policy_property(prop="user_api_rate_limit")
    def check_user_api_rate_limit(self, value: int):
        if value is not None and not isinstance(value, int):
            raise ValueError("Invalid user_api_rate_limit")

    allow_any_account = StorageGateway.connector_property("allow_any_account")

    @checks_policy_property(prop="allow_any_account")
    def check_allow_any_account(self, value: bool):
        if not isinstance(value, bool):
            raise ValueError("Invalid allow_any_account")

    @property
    def clientID(self):
        value = self.boxAppSettings["clientID"]
        if value is None:
            raise AttributeError("clientID")
        return value

    @property
    def clientSecret(self):
        value = self.boxAppSettings["clientSecret"]
        if value is None:
            raise AttributeError("clientSecret")
        return value

    @property
    def publicKeyID(self):
        return self.boxAppSettings["appAuth"]["publicKeyID"]

    @property
    def privateKey(self):
        return self.boxAppSettings["appAuth"]["privateKey"]

    @property
    def passphrase(self):
        return self.boxAppSettings["appAuth"]["passphrase"]

    def default_mapping(self, identity_data):
        if len(self.allowed_domains) == 1:
            username, domain = identity_data["username"].rsplit("@", 1)
            if domain in self.allowed_domains:
                return identity_data["username"]

    def get_user_display_name(self, username):
        if self.user_credential_required:
            return username

        user_info = self.__get_user_info(username)
        if user_info:
            return user_info["name"]

    def check_user_allowed(self, username):
        """
        Returns True if the /users API call returns a valid
        user after being filtered by the given username
        on the Box enterprise described in the given storage_gateway
        """
        if not super().check_user_allowed(username):
            return False

        if self.user_credential_required:
            return True

        user_info = self.__get_user_info(username)

        return bool(user_info)

    def __get_user_info(self, username):
        user_info = user_info_cache.get(username)

        if user_info is None:
            current_app.logger.debug(
                "About to get token by calling %s", self.get_enterprise_access_token
            )
            access_token = self.get_enterprise_access_token()

            try:
                params = {
                    "filter_term": username,
                    "fields": "id,login,name,max_upload_size",
                }
                headers = {"Authorization": "Bearer %s" % access_token}
                response = requests.get(
                    "https://api.box.com/2.0/users", params=params, headers=headers
                )

                if response.status_code >= 400 and response.status_code < 500:
                    current_app.logger.debug(
                        "Error contacting Box /2.0/user endpoint: %s %s",
                        response.status_code,
                        response.reason,
                    )
                    raise BoxConnectionError(
                        "Unable to get Box user info: %s", response.reason
                    )

                json_response = response.json()

                current_app.logger.debug(json_response)
                if json_response.get("total_count") == 1:
                    entry = json_response["entries"][0]
                    # Box is case insensitive but case preserving.
                    # Auth normalizes usernames to lowercase, but the mapping
                    # result may have come from email or elsewhere.
                    entry["login"] = entry["login"].lower()

                    if entry["login"] == username.lower():
                        user_info = entry
            except requests.exceptions.Timeout:
                raise BoxConnectionError("Timeout contacting api.box.com")

            except requests.exceptions.RequestException as e:
                raise BoxConnectionError("Unable to get Box user info: %s", str(e))

            user_info_cache.set(username, user_info, 30)

        return user_info

    def get_enterprise_access_token(self):
        """
        Either gets a token by the enterpriseID it is cached under,
        or makes a call to get a fresh token and caches it
        """
        storage_gateway = self
        enterpriseID = storage_gateway.enterpriseID

        if not enterpriseID:
            current_app.logger.error("Box config is missing enterpriseID")
            raise BoxConnectionError()

        token = cache.get(enterpriseID)
        if token is None:
            token, expires_in = self.get_new_enterprise_access_token()
            cache.set(enterpriseID, token, timeout=expires_in - 60)

        return token

    def get_user_access_token(self, username):
        """
        Look up userid by username, and use id to retrieve a cached token
        or request a new one
        """
        user_info = self.__get_user_info(username)
        userid = user_info["id"]
        token_info = user_token_cache.get(userid)
        if token_info is None:
            now = datetime.datetime.now(datetime.timezone.utc)
            token_info = {}
            token, expires_in = self.get_new_enterprise_access_token(userid)
            token_info = {
                "token": token,
                "token_expiry": now + datetime.timedelta(seconds=expires_in),
                "sub": user_info["id"],
                "email": user_info["login"],
                "name": user_info["name"],
                "max_upload": user_info["max_upload_size"],
            }
            user_token_cache.set(userid, token_info, timeout=expires_in - 300)

        return token_info

    def get_new_enterprise_access_token(self, userID=None):
        """
        Uses given storage policies to make a jwt grant
        Returns access_token, expires_in from the response
        """
        enterpriseID = self.enterpriseID
        clientID = self.clientID
        clientSecret = self.clientSecret
        publicKeyID = self.publicKeyID
        privateKey = self.privateKey
        passphrase = self.passphrase
        authentication_url = "https://api.box.com/oauth2/token"

        # error if the storage policies are misconfigured
        if not (
            enterpriseID
            and clientID
            and clientSecret
            and publicKeyID
            and privateKey
            and passphrase
        ):
            # current_app.logger.error("Box config is missing values")
            raise BoxConnectionError()

        # decrypt private key
        try:
            key = load_pem_private_key(
                data=privateKey.encode("utf8"),
                password=passphrase.encode("utf8"),
                backend=default_backend(),
            )
        except Exception as e:
            current_app.logger.error(
                "Unable to decrypt Box privateKey. Exception: {}".format(e)
            )
            raise BoxConnectionError()

        # contruct JWT assertion
        try:
            claims = {
                "iss": clientID,
                "sub": userID if userID else enterpriseID,
                "box_sub_type": "user" if userID else "enterprise",
                "aud": authentication_url,
                "jti": binascii.hexlify(os.urandom(64)).decode("utf-8"),
                "exp": int(round(time.time(), 0) + 30),
            }
            assertion = jwt.encode(
                claims, key, algorithm="RS512", headers={"kid": publicKeyID}
            )
        except Exception as e:
            current_app.logger.error(
                "Unable to encode JWT assertion. Exception: {}".format(e)
            )
            raise BoxConnectionError()

        # Make the request, parse the JSON,
        # and extract the access token
        try:
            data = {
                "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
                "assertion": assertion,
                "client_id": clientID,
                "client_secret": clientSecret,
            }
            response = requests.post(authentication_url, data=data).json()
            access_token = response.get("access_token")
            expires_in = response.get("expires_in", 0)

        except Exception as e:
            current_app.logger.error(
                "Unable to make token request to Box. Exception: {}".format(e)
            )
            raise BoxConnectionError()

        if access_token is None:
            current_app.logger.error(
                "Response has no access token: {}".format(response)
            )

            if "the 'exp' claim" in response.get("error_description", ""):
                raise ClockSkewError()
            else:
                raise BoxConnectionError()

        return access_token, expires_in

    @property
    def auth_callback(self):
        """
        URL of the auth callback that must be registered on the Box developer
        console for the application client_id.
        """
        try:
            return current_app.config["connectors"][self.CONNECTOR_ID].callback_url
        except Exception:
            raise AttributeError(
                "This storage gateway does not have a value for the auth_callback property"
            )

    @property
    def user_credential_required(self):
        return getattr(self, "client_id", None) is not None

    @property
    def load_dsi_module(self):
        return "box"

    @load_dsi_module.setter
    def load_dsi_module(self, value):
        pass

    @property
    def process_user(self):
        return "gcsweb"

    @process_user.setter
    def process_user(self, value):
        pass

    def __repr__(self):
        return "<BoxStorageGateway {}>".format(self.display_name)
