from __future__ import annotations

import requests
from cachelib import SimpleCache
from flask import current_app

import globus.manager.api.errors as errors
from globus.manager.constants import ConfigurationObjectTag

from .collection import Collection


class DropboxCollection(Collection):
    POLICY_DATA_TYPE = "dropbox_collection_policies#1.0.0"
    POLICY_DATA_TYPES = [POLICY_DATA_TYPE]

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.COLLECTION.name
            + "49b00fd6-63f1-48ae-b27f-d8af4589f876"
        )
    }
    cache = SimpleCache()

    def check_user_credential(self, credential):
        """
        Check whether the given credential can be used to access this
        collection. To do this, we'll hit the account info endpoint
        which gives us info about the user's ability to access dropbox. This
        may mark the credential as invalid, but will not commit the change.

        The result is cached for 60 seconds.
        """
        cache_key = credential.access_token
        valid = self.cache.get(cache_key)

        current_app.logger.debug("Cached result for %s is %s", credential, valid)
        if valid is None:
            resource = "https://api.dropboxapi.com/2/users/get_current_account"
            headers = {"Authorization": "Bearer {}".format(credential.access_token)}
            response = requests.post(resource, headers=headers)

            current_app.logger.debug(
                "Credential validity check: %s", response.status_code
            )
            if response.status_code >= 200 and response.status_code < 300:
                valid = True
            else:
                credential.invalid = True
                valid = errors.InvalidCredentialError(
                    "Your credentials do not have sufficient access to access this collection",
                    user_credential_id=credential.id,
                )
            self.cache.set(cache_key, valid, 60)

        if isinstance(valid, Exception):
            raise valid
        elif valid:
            return True

    @property
    def connector_id(self):
        return "49b00fd6-63f1-48ae-b27f-d8af4589f876"

    @connector_id.setter
    def connector_id(self, value):
        pass

    def __repr__(self):
        return "<DropboxCollection {}>".format(self.id)
