from __future__ import annotations

import requests
from flask import current_app

import globus.manager.api.errors as errors
from globus.manager.constants import ConfigurationObjectTag

from .oidc_user_credential import OIDCUserCredential


class DropboxUserCredential(OIDCUserCredential):
    POLICY_DATA_TYPE = "dropbox_user_credential_policies#1.0.0"
    POLICY_DATA_TYPES = [POLICY_DATA_TYPE]
    CONNECTOR_ID = "49b00fd6-63f1-48ae-b27f-d8af4589f876"

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.USER_CREDENTIAL.name + CONNECTOR_ID
        )
    }

    @property
    def connector_properties(self) -> list[str]:
        return super().connector_properties + ["root_info"]

    def __init__(
        self, appsync_data=None, doc=None, storage_gateway=None, credentials=None
    ):
        if doc:
            self.username = doc["username"]
            new_doc = {**doc}
            new_doc.pop("display_name", None)
            # Policies must be either empty (for an unprovisioned credential)
            # or passed in as a credentials object from the google callback
            policies = new_doc.get("policies", None)
            if policies is not None and policies != {
                "DATA_TYPE": self.POLICY_DATA_TYPE
            }:
                raise ValueError("policies")

            doc = new_doc
        super().__init__(
            appsync_data=appsync_data, doc=doc, storage_gateway=storage_gateway
        )

        if doc:
            patch = self._credentials_to_patch(credentials)
            patch["deleted"] = False
            patch["provisioned"] = bool(credentials)
            self._patch(**patch)

    def _credentials_to_patch(self, credentials):
        patch = super()._credentials_to_patch(credentials)
        if credentials:
            if "policies" not in patch:
                patch["policies"] = {"DATA_TYPE": self.POLICY_DATA_TYPE}
            if credentials.access_token and "root_info" not in patch["policies"]:
                patch["policies"]["root_info"] = self._get_root_info(credentials)

        return patch

    def _get_root_info(self, credential):
        root_info = None
        headers = {"Authorization": "Bearer %s" % credential.access_token}
        response = requests.post(
            "https://api.dropboxapi.com/2/users/get_current_account",
            headers=headers,
        )

        if response.status_code >= 200 and response.status_code < 300:
            json_response = response.json()
            current_app.logger.debug(json_response)
            root_info = json_response.get("root_info")
        else:
            current_app.logger.debug(
                "Error contacting Dropbox account info endpoint: %s %s",
                response.status_code,
                response.reason,
            )

        if root_info is None:
            credential.invalid = True
            raise errors.InvalidCredentialError(
                "Failure requesting Dropbox account info"
            )
        return root_info

    root_info = OIDCUserCredential.connector_property("root_info")

    @property
    def connector_id(self):
        return self.CONNECTOR_ID

    @connector_id.setter
    def connector_id(self, value):
        pass

    @property
    def token_uri(self) -> str:
        return "https://api.dropboxapi.com/oauth2/token"

    def __repr__(self):
        return "<DropboxUserCredential {}>".format(self.id)
