from __future__ import annotations

import enum
import json
import typing as t

from globus.manager.database import db

from .helpers import UTCArrowType, ensure_arrowtime


class PendingUpdate(db.Model):
    __tablename__ = "pending_updates"

    class PendingUpdateType(enum.Enum):
        DELETE = "DELETE"
        UPDATE_DOMAIN = "UPDATE_DOMAIN"
        PATCH = "PATCH"
        UPDATE_LAST_ACCESS = "UPDATE_LAST_ACCESS"

    UPDATE_DOMAIN = PendingUpdateType.UPDATE_DOMAIN
    DELETE = PendingUpdateType.DELETE
    PATCH = PendingUpdateType.PATCH
    UPDATE_LAST_ACCESS = PendingUpdateType.UPDATE_LAST_ACCESS

    # Primary key needed for ORM.
    update_id = db.Column(
        "update_id",
        db.INTEGER,
        autoincrement=True,
        primary_key=True,
        doc="ID of this pending_update",
    )
    # Allow object_id to be non-unique in this table
    object_id = db.Column(
        "object_id",
        db.TEXT,
        db.ForeignKey("configuration_object.id"),
        nullable=False,
        doc="ID of this object",
    )
    change_time = db.Column(
        UTCArrowType,
        default=db.func.now(),
        nullable=False,
        doc="Timestamp when an object change triggered a delayed update",
    )
    endpoint_serial = db.Column(
        db.TEXT,
        nullable=False,
        doc="Endpoint serial number the object change depends on",
    )
    update_type = db.Column(db.Enum(PendingUpdateType), nullable=False)
    collection_serial = db.Column(
        db.TEXT,
        nullable=True,
        doc="Mapped collection serial number the object change depends on",
    )
    pending_object = db.relationship(
        "ConfigurationObject",
        foreign_keys=[object_id],
        backref=db.backref("pending_updates", cascade="delete"),
        lazy="joined",
    )
    _data = db.Column(
        "data",
        # Note: JSON type not available until SQLite 3.9.
        db.TEXT,
        nullable=True,
        doc="PATCH document",
    )

    @db.validates("change_time")
    def _ensure_arrowtime(self, key, value):
        return ensure_arrowtime(value)

    @property
    def data(self) -> t.Optional[dict[str, t.Any]]:
        """
        Cast the JSON string to a Python object and return it.
        """
        if self._data is not None:
            return json.loads(self._data)
        return None

    @data.setter
    def data(self, value: dict[str, t.Any]) -> None:
        """
        Cast value to a JSON string and store it in the data column
        """
        self._data = json.dumps(value)
