import re
from typing import Optional, Union

import marshmallow as ma
from typing_extensions import Literal, TypedDict

from globus.manager.scripts.utils.dns_patterns import DOMAIN_NAME_PATTERN

from .data_type import Schema, data_type, data_type_schema


class ManagedDomainDict(TypedDict, total=False):
    DATA_TYPE: Literal["domain#1.0.0"]
    domain_name: str
    wildcard: bool
    certificate: str
    certificate_chain: Optional[str]
    private_key: str


class UnmanagedDomainDict(TypedDict, total=False):
    DATA_TYPE: Literal["domain#1.0.0"]
    domain_name: str
    wildcard: bool
    certificate_path: str
    certificate_chain_path: Optional[str]
    private_key_path: str


DomainDict = Union[ManagedDomainDict, UnmanagedDomainDict]


@data_type("domain#1.0.0")
class Domain_1_0_0(Schema):
    """
    Custom domain description
    """

    domain_name = ma.fields.Str(
        required=True,
        validate=[ma.validate.Regexp(DOMAIN_NAME_PATTERN), ma.validate.Length(max=253)],
        description="Domain name",
    )
    wildcard = ma.fields.Boolean(
        required=True,
        description="""
        Flag indicating whether this is a wildcard domain or not.

        When setting a custom domain for a mapped collection, the domain
        may optionally be a wildcard domain. If it is a wildcard domain,
        the guest collections will be created as subdomains of the mapped
        collection domain; if not, guest collections will be created as
        subdomains of the endpoint domain.
        """,
    )
    certificate = ma.fields.Str(
        required=False,
        allow_none=True,
        validate=ma.validate.Regexp(
            "^(.*\n)?-----BEGIN CERTIFICATE-----.*", re.MULTILINE
        ),
        description="PEM-Encoded X.509 certificate for this domain",
    )
    certificate_chain = ma.fields.Str(
        required=False,
        allow_none=True,
        validate=ma.validate.Regexp(
            "^(.*\n)?-----BEGIN CERTIFICATE-----.*", re.MULTILINE
        ),
        description="""
        PEM-Encoded X.509 certificate chain for this domain. Only needed if
        there are intermediate certificates that must also be sent to
        clients to allow them to verify the certificate.
        """,
    )
    private_key = ma.fields.Str(
        required=False,
        allow_none=True,
        validate=ma.validate.Regexp(
            "^(.*\n)?-----BEGIN PRIVATE KEY-----.*", re.MULTILINE
        ),
        description="PEM-Encoded private key for the certificate",
    )
    certificate_path = ma.fields.Str(
        required=False,
        allow_none=True,
        validate=ma.validate.Regexp("^/.*"),
        description="""
        Path to a file containing the X.509 certificate for this domain.
        This file path must contain a valid certificate and be present on
        each data transfer node.
        """,
    )
    certificate_chain_path = ma.fields.Str(
        required=False,
        allow_none=True,
        validate=ma.validate.Regexp("^/.*"),
        description="""
        Path to a file containing the X.509 certificate chain for this
        domain. This file path must contain a sequence of valid
        certificate and be present on each data transfer node.
        """,
    )
    private_key_path = ma.fields.Str(
        required=False,
        allow_none=True,
        validate=ma.validate.Regexp("^/.*"),
        description="""
        Path to a file containing the private key for this domain. This
        file path must contain a valid key and be present on each data
        transfer node.
        """,
    )

    def dump(self, *args, **kwargs) -> DomainDict:
        return super().dump(*args, **kwargs)

    def load(self, *args, **kwargs) -> DomainDict:
        return super().load(*args, **kwargs)

    private = [
        "certificate_chain",
        "private_key",
        "certificate_path",
        "certificate_chain_path",
        "private_key_path",
    ]

    @ma.validates_schema
    def validate_managed_or_unmanaged(self, data, **kwargs):
        # A domain can be either managed or unmanaged.
        # For a managed domain:
        # - both the certificate and private_key must be present and be a valid
        #   certificate and key combination
        # - certificate_chain may be set to include a list of other certs
        # - both certificate_path and private_key_path must not be set or must be
        #   set to null.
        # For an unmanaged domain:
        # - both the certificate_path and private_key_path must be present and
        #   point to existing files.
        # - certificate_chain_path may optionally be set and if it is non-null
        #   must point to a valid certificate chain file.
        # - the certificate and private_key values must not be set or must be
        #   set to null.
        certificate = data.get("certificate")
        certificate_chain = data.get("certificate_chain")
        private_key = data.get("private_key")

        certificate_path = data.get("certificate_path")
        certificate_chain_path = data.get("certificate_chain_path")
        private_key_path = data.get("private_key_path")

        managed = certificate and private_key
        unmanaged = certificate_path and private_key_path

        if (
            (managed and unmanaged)
            or not (managed or unmanaged)
            or (managed and certificate_chain_path)
            or (unmanaged and certificate_chain)
        ):
            raise ma.ValidationError(
                "A domain#1.0.0 document must contain both certificate_chain and "
                "private_key or both certificate_path and private_key_path"
            )


Domain = data_type_schema.Domain
