import ipaddress
import typing as t

import marshmallow as ma

from globus.manager.constants import NodeStatus

from ..enum_field import EnumField
from .data_type import Schema, data_type, data_type_schema


@data_type("node#1.0.0")
class Node_1_0_0(Schema):
    __doc__ = """
    Services for Globus Connect Server endpoints may be deployed on multiple
    different physical resources, referred to as data transfer nodes. Each node
    may have one or more IP addresses, TCP incoming and outgoing port ranges,
    and a *status* value indicating whether it is configured to actively
    respond to requests or is in maintenance mode.
    """

    id = ma.fields.UUID(
        description="""
        Unique id string this node. This is system generated and may not
        be included in create requests.
        """
    )
    ip_addresses = ma.fields.List(
        ma.fields.IP,
        required=True,
        description="List of IP addresses for this node",
    )

    incoming_port_range = ma.fields.List(
        ma.fields.Int(validate=[ma.validate.Range(min=0, max=2**16 - 1)]),
        validate=[ma.validate.Length(equal=2)],
        required=False,
        description="Allowed port range for incoming TCP data connections",
    )

    outgoing_port_range = ma.fields.List(
        ma.fields.Int(validate=[ma.validate.Range(min=0, max=2**16 - 1)]),
        validate=[ma.validate.Length(equal=2)],
        required=False,
        description="Port range used as the source for outgoing TCP data connections",
    )

    status = EnumField(
        NodeStatus,
        required=True,
        description="""
        Current status of the Node. If a Node is marked *inactive*, it will
        be removed from the DNS entries for this endpoint and will return
        an error on any attempt to use the Manager API or attempt a
        Transfer using this node.
        """,
    )


@data_type("node#1.1.0")
class Node_1_1_0(Node_1_0_0):
    __doc__ = (
        t.cast(str, Node_1_0_0.__doc__)
        + """
    Version 1.1.0 adds support for setting the data interface on a node.
    """
    )
    data_interface = ma.fields.IPv4(
        required=False,
        allow_none=True,
        description="IP address on which this node listens for data transfers",
    )

    @ma.validates("data_interface")
    def validate_data_interface(self, data, **kwargs):
        if data:
            try:
                ipaddress.ip_address(data)
            except Exception as e:
                raise ma.ValidationError(str(e))


@data_type("node#1.2.0")
class Node_1_2_0(Node_1_1_0):
    __doc__ = (
        t.cast(str, Node_1_1_0.__doc__)
        + """
    Version 1.2.0 adds support for setting an IPv6 data interface on a node.
    """
    )
    data_interface6 = ma.fields.IPv6(
        required=False,
        allow_none=True,
        description="IPv6 address on which this node listens for data transfers",
    )


Node = data_type_schema.Node
