import uuid
from typing import List, Optional

import marshmallow as ma
from typing_extensions import Literal, TypedDict

from .data_type import Schema, data_type, data_type_schema
from .path_restrictions import PathRestrictions, PathRestrictionsDict


class SharingPolicyDumpDict(TypedDict, total=False):
    DATA_TYPE: Literal["sharing_policy#1.0.0"]
    id: str
    collection_id: str
    sharing_restrict_paths: PathRestrictionsDict
    users: Optional[List[str]]


class SharingPolicyLoadDict(TypedDict, total=False):
    DATA_TYPE: Literal["sharing_policy#1.0.0"]
    id: uuid.UUID
    collection_id: uuid.UUID
    sharing_restrict_paths: PathRestrictionsDict
    users: Optional[List[str]]


@data_type("sharing_policy#1.0.0")
class SharingPolicy_1_0_0(Schema):
    """
    Sharing policies for a mapped collection.

    This document type allows endpoint and collection administrators to
    optionally constrain sharing path policies for particular users. The
    **sharing_restrict_paths** property has a similar meaning to that of the
    **sharing_restrict_paths** in the collection document; however, it is in
    effect only for specific users.

    If the **users** property is null, then the restriction applies to all
    users. If it is non-null, then this restriction applies only to accounts
    which have been mapped to the enumerated storage gateway user accounts.

    Multiple sharing policies can be defined for a mapped collection.  When a
    guest collection is created or accessed, only the policies relevant to the
    user which created the account are enforced.
    """

    id = ma.fields.UUID(description="Unique id for this sharing policy")
    collection_id = ma.fields.UUID(
        required=True,
        description="Id of the mapped collection which this policy is associated with",
    )
    sharing_restrict_paths = ma.fields.Nested(
        PathRestrictions,
        required=True,
        description="""
        Restrictions on which paths may be shared in guest collections
        related to this mapped collection. These paths are relative to the
        root_path property of the mapped collection.
        """,
    )
    users = ma.fields.List(
        ma.fields.String,
        description="""
        List of local user accounts that this policy applies to. If omitted
        or null, this restriction applies to all local user accounts.
        """,
    )

    def load(self, *args, **kwargs) -> SharingPolicyLoadDict:
        return super().load(*args, **kwargs)


SharingPolicy = data_type_schema.SharingPolicy
