import marshmallow as ma

from ..api import (
    PolicySchema,
    collection_policy,
    data_type_schema,
    storage_gateway_policy,
    user_credential_policy,
)
from ..cleanup_none_mixin import CleanupNoneMixin

POSIX_CONNECTOR_ID = "145812c8-decc-41f1-83cf-bb2a85a2a70b"


@storage_gateway_policy("posix_storage_policies#1.0.0")
class PosixStoragePolicies_1_0_0(PolicySchema, CleanupNoneMixin):
    """
    Connector-specific storage gateway policies for the POSIX connector.
    """

    groups_allow = ma.fields.List(
        ma.fields.Str,
        allow_none=True,
        description="""
        List of POSIX group names allowed to access this storage gateway
        """,
    )
    groups_deny = ma.fields.List(
        ma.fields.Str,
        allow_none=True,
        description="""
        List of POSIX group names denied access this storage gateway
        """,
    )

    private = ["groups_allow", "groups_deny"]
    connector_id = POSIX_CONNECTOR_ID


@collection_policy("posix_collection_policies#1.0.0")
class PosixCollectionPolicies_1_0_0(PolicySchema):
    """
    Connector-specific collection policies for the POSIX connector
    """

    connector_id = POSIX_CONNECTOR_ID


@collection_policy("posix_collection_policies#1.1.0")
class PosixCollectionPolicies_1_1_0(PosixCollectionPolicies_1_0_0):
    __doc__ = f"""
    {getattr(PosixCollectionPolicies_1_0_0, '__doc__', '').lstrip()}

    Version 1.1.0 of the posix_collection_policies document adds the
    **sharing_groups_allow**, and **sharing_groups_deny** properties.
    """

    sharing_groups_allow = ma.fields.List(
        ma.fields.Str,
        allow_none=True,
        description="""
        List of POSIX group names allowed to create shares on this collection
        """,
    )
    sharing_groups_deny = ma.fields.List(
        ma.fields.Str,
        allow_none=True,
        description="""
        List of POSIX group names denied access to create shares on this
        collection.
        """,
    )
    private = ["sharing_groups_allow", "sharing_groups_deny"]
    connector_id = POSIX_CONNECTOR_ID


@user_credential_policy("posix_user_credential_policies#1.0.0")
class PosixUserCredentialPolicies_1_0_0(PolicySchema):
    """
    Connector-specific user credential policies for the POSIX connector
    """

    connector_id = POSIX_CONNECTOR_ID


PosixCollectionPolicies = data_type_schema.PosixCollectionPolicies
PosixStoragePolicies = data_type_schema.PosixStoragePolicies
PosixUserCredentialPolicies = data_type_schema.PosixUserCredentialPolicies
