"""
Helpers for interacting with Globus Search
"""
import logging
import time

from flask import current_app

import globus_sdk

from globus.manager.constants import SEARCH_GCS_INDEX_ID

log = logging.getLogger(__name__)


def suppress_search_errors(func, *args, **kwargs):
    try:
        return func(*args, **kwargs)
    except globus_sdk.SearchAPIError:
        log.exception("suppressed Search API error")
        return None


def submit_single_search_document(subject, document, visible_to=None):
    """
    Common wrapper for submitting documents (one at a time) to Search. Returns
    the Search API response.
    Task submission may fail in certain cases -- e.g. schema validation fails
    -- in which case this function will raise a globus_sdk.SearchAPIError

    :param subject: The ID of the document being sent to Search. This will be
                    validated to ensure that it starts with the ClientID and
                    ends with one of the valid suffixes.
    :param document: A dictionary containing the data to send to Search. This
                     is the part which will be evaluated against the JSONSchema
                     in Search.
    :param visible_to: A list of principals to use for the visible_to setting
                       on this document. Defaults to ["public"]
    """
    if visible_to is None:
        visible_to = ["public"]

    # produce a GIngest containing a GMetaEntry containing the document
    # GIngest is used to let Search take a single item or an array and may be
    # used to parameterize ingest in the future:
    #   https://docs.globus.org/api/search/ingest/#gingest
    #
    # GMetaEntry combines subject, visible_to, and "content" (the input
    # document). Its fields are documented here:
    #   https://docs.globus.org/api/search/ingest/#gmetaentry
    wrapped_document = {
        "ingest_type": "GMetaEntry",
        "ingest_data": {
            "subject": subject,
            "visible_to": visible_to,
            "content": document,
        },
    }
    return current_app.search_client.ingest(SEARCH_GCS_INDEX_ID, wrapped_document)


def submit_multiple_search_documents(subjects_and_documents, visible_to=None):
    """
    Wrapper for submitting documents in bulk to Search. Returns the Search API
    response.
    Task submission may fail in certain cases, in which case this function will
    raise a globus_sdk.SearchAPIError

    :param subjects_and_documents: A list of tuples of (subject, document). The
                                   meanings of "subject" and "document" are the
                                   same as in submit_single_search_document
    :param visible_to: A list of principals to use for the visible_to setting
                       on this document. Defaults to ["public"]
    """
    if visible_to is None:
        visible_to = ["public"]

    # produce a GIngest containing a GMetaList containing the documents
    # GIngest is used to let Search take a single item or an array and may be
    # used to parameterize ingest in the future:
    #   https://docs.globus.org/api/search/ingest/#gingest
    #
    # GMetaList is an array of GMetaEntry documents.
    # Its fields are documented here:
    #   https://docs.globus.org/api/search/ingest/#gmetalist
    wrapped_document = {
        "ingest_type": "GMetaList",
        "ingest_data": {
            "gmeta": [
                {"subject": subject, "visible_to": visible_to, "content": document}
                for (subject, document) in subjects_and_documents
            ]
        },
    }
    return current_app.search_client.ingest(SEARCH_GCS_INDEX_ID, wrapped_document)


def delete_search_document(subject):
    """
    Delete a document from Search by ID ("subject").
    Catches and discards 404s (if the document is double-deleted, for example)
    but otherwise raises any errors encountered.

    The subject will be checked against the ClientID in Search to ensure that,
    by policy, this Endpoint is allowed to delete the document.

    :param subject: The subject of the document to delete
    """
    try:
        current_app.search_client.delete_subject(SEARCH_GCS_INDEX_ID, subject)
    except globus_sdk.SearchAPIError as e:
        if e.http_status != 404:
            raise e


def wait_for_search_task(task_id, retries=3):
    """
    Wait with exponential backoff for a Search task to succeed.
    If it fails or does not succeed within the timeline, returns False.
    If it succeeds, returns True.

    :param task_id: The ID of the task to wait for. It must have been submitted
                    by the same Endpoint which is now querying for it.
    :param retries: How many times to retry before giving up. (retries=1 means
                    "try once and retry once").
    """
    trynum = 0
    while trynum < retries + 1:
        res = current_app.search_client.get_task(task_id)
        if res["state"] == "SUCCESS":
            return True
        elif res["state"] == "FAILED":
            return False
        else:
            time.sleep(0.5 * (2**trynum))
            trynum += 1
    return False
