import os

from jwcrypto.jwk import JWK


class GridFTPKeyManager:
    """
    Class that periodically updates the authentication key used by the GridFTP
    for performing login checks to the GCS Manager API.
    """

    # Every 10 minutes, rotate the gridftp login key
    UPDATE_PERIOD = 60 * 10
    PRIORITY = 1

    def __init__(self, scheduler, app):
        self.scheduler = scheduler
        self.app = app
        self.scheduler.enter(0, self.PRIORITY, self.update_gridftp_key)
        self.old_key = None
        self.gridftp_key_path = os.path.expanduser("~gcsweb/gridftp-key")
        self.gridftp_key_path_new = self.gridftp_key_path + ".new"
        self.gridftp_key_path_old = self.gridftp_key_path + ".old"

    def update_gridftp_key(self):
        """
        Update the gridftp keys on disk, by copying the existing key
        to the old key path ane writing the new key to a temp file name
        and then renaming it to the well-known name.
        """
        self.scheduler.enter(self.UPDATE_PERIOD, self.PRIORITY, self.update_gridftp_key)

        # If there is an existing key, keep 1 previous key per run of
        # this script
        if self.old_key:
            self._write_key(self.gridftp_key_path_old, self.old_key)
        else:
            try:
                os.remove(self.gridftp_key_path_old)
            except IOError:
                pass

        # Generate new key, write it to a temporary name, and then rename it
        new_key = self.generate_key()
        self._write_key(self.gridftp_key_path_new, new_key)
        os.rename(self.gridftp_key_path_new, self.gridftp_key_path)

        # Track the new key, so we can stash it in the old key path
        self.old_key = new_key

    def generate_key(self):
        gridftp_key = JWK.generate(kty="oct")
        return gridftp_key.get_op_key()

    def _write_key(self, path, key):
        """
        Write the key data to the given path, with mode 0600
        """
        with open(os.open(path, os.O_CREAT | os.O_WRONLY, 0o0600), "w") as f:
            f.write(key)
