import glob
import os
import re
from datetime import datetime, timezone

import click
from xopen import xopen

from gcs_cli.parsing.command_decorator import globus_command

from .db import Operation, Session, create_app, db

DEFAULT_LOG_PATH = "/var/log/gridftp-audit"
DEFAULT_DB_PATH = "~/.globus/audit-logs.db"


@globus_command(
    "load",
    help="Load audit logs into database for searching",
    disable_options=["format", "explicit_host"],
)
@click.option(
    "--verbose",
    "-v",
    help="Show verbose output.",
    is_flag=True,
)
@click.option(
    "--log-dir",
    "-l",
    help="Directory containing GridFTP audit logs.",
    type=click.Path(exists=True),
    metavar="PATH",
    show_default=True,
    default=DEFAULT_LOG_PATH,
)
@click.option(
    "--db-path",
    "-d",
    help="Path where database will be created. Any existing database at this path will be removed.",
    type=click.Path(dir_okay=False),
    metavar="PATH",
    show_default=True,
    default=DEFAULT_DB_PATH,
)
def audit_load(log_dir, db_path, verbose):
    vprint = print if verbose else lambda *a, **k: None
    db_path = os.path.expanduser(db_path)
    try:
        os.remove(db_path)
        vprint("Removed {}".format(db_path))
    except FileNotFoundError:
        pass
    vprint("Creating {}".format(db_path))
    os.makedirs(os.path.dirname(db_path), exist_ok=True)
    app = create_app(db_path)
    with app.app_context():
        print("Loading logs from {}...".format(log_dir))
        active_log = []
        if log_dir == DEFAULT_LOG_PATH:
            active_log = [DEFAULT_LOG_PATH + ".log"]
        for logfile in (
            sorted(glob.glob(os.path.join(log_dir, "gridftp-audit.log-*"))) + active_log
        ):
            vprint("  reading {}".format(logfile))
            with xopen(logfile, mode="rt") as file:
                _cnt = 0
                for _cnt, line in enumerate(file):
                    ent = dict(
                        [
                            (k[0], k[2] if k[2] else k[1])
                            for k in re.findall(r'(\w*)=([^"\s]\S*|"(.+?)")', line)
                        ]
                    )
                    load_entry(ent)
                vprint("   read {} lines.".format(_cnt))
        print("Done.")


def load_entry(ent):
    ts = datetime.strptime(ent.get("ts"), "%Y-%m-%dT%H:%M:%SZ").replace(
        tzinfo=timezone.utc
    )
    if ent.get("t") == "S":
        record = Session(
            id=ent.get("ref"),
            ts=ts,
            client_addr=ent.get("c_ip"),
            auth_type=ent.get("auth"),
            process_user=ent.get("p_usr"),
            storage_user=ent.get("s_usr"),
            root_path=ent.get("r"),
            task_id=ent.get("g_tid"),
            collection_id=ent.get("g_eid"),
            identity_id=ent.get("g_uid"),
        )
        db.session.add(record)

    elif ent.get("t") == "O":
        record = Operation.query.filter(
            Operation.session_id == ent.get("ref"),
            Operation.op == ent.get("op"),
            Operation.path == ent.get("c_path"),
            Operation.end_ts == datetime.min,
        ).first()
        if record:
            record.count += 1
        else:
            record = Operation(
                session_id=ent.get("ref"),
                start_ts=ts,
                end_ts=datetime.min,
                op=ent.get("op"),
                path=ent.get("c_path"),
                result=ent.get("res", "ok"),
                count=1,
            )
            db.session.add(record)

    elif ent.get("t") == "E":
        record = Operation.query.filter(
            Operation.session_id == ent.get("ref"),
            Operation.op == ent.get("op"),
            Operation.path == ent.get("c_path"),
            Operation.end_ts == datetime.min,
        ).first()
        if record.op in ("STOR", "RETR"):
            record.end_ts = ts
            record.peer_addr = ent.get("d_ip")
            record.length = ent.get("len")
        record.result = ent.get("res")

    db.session.commit()
