import json
import os
import textwrap

import click

import globus.manager.api.models as models
from globus.manager.api.app import make_app
from globus.manager.api.blueprints.utils import update_object
from globus.manager.scripts.utils.setup_utils import load_keychain
from globus.manager.utils.keychain import NodeKey


def certificate_renew() -> None:
    ctx = click.get_current_context()
    if os.getuid() != 0:
        click.echo("This command must be run as the root user", err=True)
        ctx.exit(1)

    click.echo(
        "\n".join(
            textwrap.wrap(
                "This script renews the certificate for a GCS endpoint. "
                "The renewal uses a DNS-based challenge to verify domain "
                "ownership. The process may take a few minutes."
            )
        )
    )

    app = make_app(app_config="globus.manager.api.config.certificate_cli")
    models.db.init_app(app)

    with app.app_context():
        models.db.create_all()
        models.db.session.commit()

        # Needed by AppSync
        for conf in models.Configuration.query:
            if conf.name == "node_key":
                value = NodeKey(**json.loads(conf.value))
            else:
                value = conf.value
            app.config[conf.name.upper()] = value

        # Make sure `node setup` has been run
        if "CLIENT_ID" not in app.config or "CLIENT_SECRET" not in app.config:
            click.echo("This node has not been set up.", err=True)
            ctx.exit(1)

        # Load the keychain
        load_keychain()
        try:
            acme_account = models.AcmeAccount.query.one()

            def acquire_lease_update(account):
                account.acquire_lease()

            update_object(acme_account, acquire_lease_update)
        except Exception as e:
            click.echo(
                "\n".join(
                    textwrap.wrap(
                        "Unable to acquire renewal lock, another node may be "
                        "renewing the certificate currently: {}".format(e)
                    )
                ),
                err=True,
            )
            ctx.exit(1)

        try:
            certificate = models.LetsEncryptCertificate.query.one()

            def renew_cert(cert):
                cert.renew()

            update_object(certificate, renew_cert)
        except Exception as e:
            click.echo("Unable to renew certificate: {}".format(e), err=True)
            ctx.exit(1)

        try:

            def release_lease_update(account):
                account.release_lease()

            update_object(acme_account, release_lease_update)
        except Exception as e:
            # Non fatal, it will expire
            click.echo("Warning: unable to release lock: {}".format(str(e)))
            pass
        click.echo("Certificate renewed successfully.")
