from typing import Tuple

import click

from gcs_cli.parsing.command_decorator import globus_command

from globus.manager.scripts.utils.dns_patterns import verify_domain_name_option
from globus.manager.scripts.utils.email_patterns import verify_email_option


@globus_command(
    name="register",
    short_help="Register an existing OIDC server for use by this endpoint",
    disable_options=["format", "explicit_host"],
    help="""\
The OIDC register command registers an existing OIDC server as a Globus
identity proivder, for use with this endpoint.

Once configured, you will be able to authenticate users from the associated domain.

== EXAMPLE

globus-connect-server oidc register

""",
)
@click.option("--display-name", help="Display name for the OIDC server", required=True)
@click.option(
    "--discovery-url",
    metavar="URL",
    help=(
        "The OpenID Connect discovery URL for the server.  The domain of the url "
        "will be the main domain of the IdP and must be asserted by the OIDC server. "
        "Example: https://id.example.com/.well-known/openid-configuration"
    ),
    required=True,
)
@click.option(
    "--domain",
    metavar="DOMAIN",
    help=(
        "An alternate domain asserted by the OIDC server.  It is not necessary to "
        "pass the domain from the 'discovery-url'.  May be passed multiple times."
    ),
    callback=verify_domain_name_option,
    multiple=True,
)
@click.option(
    "--client-id",
    metavar="CLIENT-ID",
    help=(
        "You must configure Globus Auth as a client on your OIDC Server.  This is the "
        "Client ID of that that client configuration, which Globus will use to "
        "authenticate with your OIDC server. "
        "This is not the Globus client-id of your endpoint. "
    ),
    required=True,
)
@click.option(
    "--client-secret",
    metavar="SECRET",
    help=(
        "You must configure Globus Auth as a client on your OIDC Server.  This is the "
        "Client secret for the Client ID that Globus will use to authenticate "
        "with your OIDC server.  This is not the Globus client-secret of your endpoint."
    ),
    required=True,
)
@click.option(
    "--username-claim",
    help=(
        "Identity provider claim that maps to the user name. In most cases this "
        "should be 'preferred_username'."
    ),
    required=True,
)
@click.option(
    "--id-claim",
    help=(
        "Identity provider claim that maps to the immutable ID. In most cases "
        "this should be 'sub'."
    ),
    required=True,
)
@click.option(
    "--support-contact", help="Support contact name for the OIDC server", required=True
)
@click.option(
    "--support-email",
    help="Support contact email address for the OIDC server",
    required=True,
    callback=verify_email_option,
)
@click.pass_context
def oidc_register(
    ctx,
    display_name: str,
    discovery_url: str,
    domain: Tuple[str],
    support_contact: str,
    support_email: str,
    client_id: str,
    client_secret: str,
    username_claim: str,
    id_claim: str,
) -> None:
    from .oidc_register import oidc_register as oir

    oir(
        ctx,
        display_name,
        discovery_url,
        domain,
        support_contact,
        support_email,
        client_id,
        client_secret,
        username_claim,
        id_claim,
    )
