"""
Systemd Service Manager
"""

from __future__ import absolute_import

from subprocess import PIPE, Popen


def disable(service_name, periodic=False, path=False, socket=False):
    """
    Disable the named service so that it will be started automatically
    on system boot.
    """
    units = []

    if periodic:
        units.append(f"{service_name}.timer")
    if path:
        units.append(f"{service_name}.path")
    if socket:
        units.append(f"{service_name}.socket")
    units.append(f"{service_name}.service")

    args = ["systemctl", "-q", "disable"] + units
    disabler = Popen(args, stdin=None)
    disabler.communicate()
    return disabler.returncode == 0


def enable(service_name, periodic=False, path=False, socket=False):
    """
    Enable the named service so that it will not be started automatically
    at system boot.
    """
    units = []
    if periodic:
        units.append(f"{service_name}.timer")
    if path:
        units.append(f"{service_name}.path")
    if socket:
        units.append(f"{service_name}.socket")
    units.append(f"{service_name}.service")

    args = ["systemctl", "-q", "enable"] + units
    enabler = Popen(args, stdin=None)
    enabler.communicate()
    return enabler.returncode == 0


def is_enabled(service_name: str) -> bool:
    """
    Return True if the named service is enabled.
    """
    args = ["systemctl", "-q", "enable", service_name]
    enabler = Popen(args, stdin=None)
    enabler.communicate()
    return enabler.returncode == 0


def start(service_name, periodic=False, path=False, socket=False):
    """
    Start the named service.
    """
    units = []
    if periodic:
        units.append(f"{service_name}.timer")
    if path:
        units.append(f"{service_name}.path")
    if socket:
        units.append(f"{service_name}.socket")
    if not periodic or path or socket:
        units.append(f"{service_name}.service")

    args = ["systemctl", "-q", "start"] + units
    starter = Popen(args, stdin=None)
    starter.communicate()
    return starter.returncode == 0


def stop(service_name, periodic=False, path=False, socket=False):
    """
    Stop the named service.
    """
    units = []
    if periodic:
        units.append(f"{service_name}.timer")
    if path:
        units.append(f"{service_name}.path")
    if socket:
        units.append(f"{service_name}.socket")
    units.append(f"{service_name}.service")
    args = ["systemctl", "-q", "stop"] + units
    stopper = Popen(args, stdin=None)
    stopper.communicate()
    return stopper.returncode == 0


def status(service_name):
    """
    Return the status of the named service. Returns True if the service
    is running
    """
    status = Popen(
        ["systemctl", "-q", "status", service_name + ".service"],
        stdin=None,
        stdout=PIPE,
        stderr=PIPE,
    )
    status.communicate()

    return status.returncode == 0


def restart(service_name, periodic=False, path=False, socket=False):
    """
    Restart the named service.
    """
    units = []
    if periodic:
        units.append(f"{service_name}.timer")
    if path:
        units.append(f"{service_name}.path")
    if socket:
        units.append(f"{service_name}.socket")
    units.append(f"{service_name}.service")
    args = ["systemctl", "-q", "restart"] + units
    restarter = Popen(args, stdin=None)
    restarter.communicate()

    return restarter.returncode == 0


def reload(service_name):
    """
    Reload the named service's configuration.
    """
    args = ["systemctl", "-q", "reload", service_name + ".service"]
    reloader = Popen(args, stdin=None)
    reloader.communicate()

    return reloader.returncode == 0
